from pathlib import Path
from typing import Dict
from .base import SchematisationChildWrapper, RevisionChildWrapper
from threedi_api_client.openapi.api import V3BetaApi
from threedi_api_client.openapi.models import (
    SchematisationRevision,
    SqliteFileUpload,
    RevisionRaster,
    Commit,
    Upload,
    Sqlite,
)
from .waitfor import WaitForModel


class RevisionWrapper(SchematisationChildWrapper):
    model = SchematisationRevision
    api_path: str = "revisions"
    scenario_name = "revision"


class WaitForSqliteUploadWrapper(WaitForModel):
    model = Sqlite
    scenario_name = "waitforsqliteupload"
    websocket_event_type = "event"

    def matches(self, websocket_instance):
        if (
            isinstance(websocket_instance, self.model)
            and websocket_instance.revision_id == self.instance.revision_id
        ):
            return websocket_instance.file["state"] == "uploaded"


class WaitForRasterUploadWrapper(WaitForModel):
    model = RevisionRaster
    scenario_name = "waitforrevisionrastereupload"
    websocket_event_type = "event"

    def matches(self, websocket_instance):
        if (
            isinstance(websocket_instance, self.model)
            and websocket_instance.id == self.instance.id
        ):
            return websocket_instance.file["state"] == "uploaded"


class SqliteWrapper(RevisionChildWrapper):
    model = SqliteFileUpload
    api_path: str = "revisions_sqlite_upload"
    scenario_name = "sqlite"
    filepath: Path = None

    def resolve_func_name(self, suffix: str):
        if suffix == "_create":
            suffix = ""
        return super().resolve_func_name(suffix)

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        wait_for_upload = WaitForSqliteUploadWrapper(
            data={"revision_id": self.revision.id},
            api_client=self._api_client,
        )
        return [wait_for_upload]


class RasterWrapper(RevisionChildWrapper):
    model = RevisionRaster
    api_path: str = "revisions_rasters"
    scenario_name = "raster"
    filepath: Path = None
    filename: str = None

    def initialize_instance(self, data: Dict):
        self.filename = data.pop("filename")
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    def save(self):
        super().save(upload_file=False)

        # Use raster to create raster upload
        api_client: V3BetaApi = self._api_client
        upload: Upload = Upload(filename=self.filename)
        upload = api_client.schematisations_revisions_rasters_upload(
            self.instance.id, self.revision.id, self.schematisation.id, upload
        )

        # Upload raster
        self._upload_file(upload.put_url)

    @property
    def extra_steps(self):
        wait_for_upload = WaitForRasterUploadWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
        )
        return [wait_for_upload]


class CommitWrapper(RevisionChildWrapper):
    model = Commit
    api_path: str = "revisions_commit"
    scenario_name = "commit"

    def resolve_func_name(self, suffix: str):
        if suffix == "_create":
            suffix = ""
        return super().resolve_func_name(suffix)


WRAPPERS = [
    RevisionWrapper,
    SqliteWrapper,
    RasterWrapper,
    CommitWrapper,
    WaitForSqliteUploadWrapper,
    WaitForRasterUploadWrapper,
]



from pathlib import Path
from sqlalchemy.orm import Query
from threedi_schema import models
from threedi_schema import ThreediDatabase
from typing import Dict
from typing import Optional


RASTERS = [
    "dem_file",
    "frict_coef_file",
    "interception_file",
    "porosity_file",
    "hydraulic_conductivity_file",
    "infiltration_rate_file",
    "max_infiltration_capacity_file",
    "groundwater_impervious_layer_level_file",
    "phreatic_storage_capacity_file",
    "initial_infiltration_rate_file",
    "equilibrium_infiltration_rate_file",
    "infiltration_decay_period_file",
    "groundwater_hydro_connectivity_file",
    "initial_waterlevel_file",
    "initial_groundwater_level_file",
]


__all__ = ["ModelDB"]


MIN_SQLITE_VERSION = 219


class ModelDB:
    """Interface to sqlite of a model."""

    def __init__(
        self,
        sqlite_path: Path,
        global_settings_id: Optional[int] = None,
        upgrade: bool = False,
    ):
        if not sqlite_path.exists():
            raise ValueError(f"Sqlite path {sqlite_path} does not exist.")

        self.sqlite_path = sqlite_path
        self.database = ThreediDatabase(self.sqlite_path.as_posix())

        version = self.get_version()
        if version < MIN_SQLITE_VERSION:
            if upgrade:
                self.upgrade()
            else:
                raise ValueError(f"Too old sqlite version {version}.")

        if global_settings_id:
            self.global_settings_id = global_settings_id
        else:
            try:
                session = self.database.get_session()
                self.global_settings_id, self.global_settings_name = Query(
                    [models.GlobalSetting.id, models.GlobalSetting.name]
                ).with_session(session)[0]
            finally:
                session.close()

    def get_version(self) -> int:
        # check version
        return self.database.schema.get_version()

    def upgrade(self):
        self.database.schema.upgrade()

    def get_raster_filepaths(self, base_path: Path) -> Dict:
        """Get all raster paths from sqlite based on global_settings id."""

        raster_filepaths = {}
        settings_map = {
            "interflow_settings_id": models.Interflow,
            "simple_infiltration_settings_id": models.SimpleInfiltration,
            "groundwater_settings_id": models.GroundWater,
        }

        try:
            session = self.database.get_session()
            glob_settings = (
                Query(models.GlobalSetting)
                .with_session(session)
                .get(self.global_settings_id)
            )
            settings = [glob_settings]
            for k, v in settings_map.items():
                settings_id = getattr(glob_settings, k)
                if settings_id:
                    settings.append(Query(v).with_session(session).get(settings_id))

            for setting in settings:
                for raster in RASTERS:
                    try:
                        raster_path = getattr(setting, raster)
                        if raster_path:
                            raster_filepaths[raster] = base_path / Path(
                                raster_path.replace("\\", "/")
                            )
                    except AttributeError:
                        pass

            if glob_settings.initial_waterlevel_file:
                raster_filepaths["initial_waterlevel_file"] = base_path / Path(
                    glob_settings.initial_waterlevel_file.replace("\\", "/")
                )
            if glob_settings.initial_groundwater_level_file:
                raster_filepaths["initial_groundwater_level_file"] = base_path / Path(
                    glob_settings.initial_groundwater_level_file.replace("\\", "/")
                )

        finally:
            session.close()


        return raster_filepaths


from typing import Dict
from pathlib import Path
from threedi_api_client.openapi.models.upload_event_file import UploadEventFile
from .base import EventWrapper
from threedi_api_client.openapi.models import (
    TimedStructureControl,
    TableStructureControl,
    MemoryStructureControl,
    FileStructureControl,
)
from .waitfor import (
    WaitForEventValidation,
    match_validated_event,
)


class WaitForTimedStructureControlWrapper(WaitForEventValidation):
    model = TimedStructureControl
    scenario_name = "waitfortimedstructurecontrol"

    def matches(self, websocket_instance):
        return match_validated_event(websocket_instance, self)


class TimedStructureControlWrapper(EventWrapper):
    model = TimedStructureControl
    api_path: str = "structure_control_timed"
    scenario_name = "structurecontroltimed"

    @property
    def extra_steps(self):
        wait_for_validation = WaitForTimedStructureControlWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


class WaitForTableStructureControlWrapper(WaitForEventValidation):
    model = TableStructureControl
    scenario_name = "waitfortablestructurecontrol"

    def matches(self, websocket_instance):
        return match_validated_event(websocket_instance, self)


class TableStructureControlWrapper(EventWrapper):
    model = TableStructureControl
    api_path: str = "structure_control_table"
    scenario_name = "structurecontroltable"

    @property
    def extra_steps(self):
        wait_for_validation = WaitForTableStructureControlWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


class WaitForMemoryStructureControlWrapper(WaitForEventValidation):
    model = MemoryStructureControl
    scenario_name = "waitformemorystructurecontrol"

    def matches(self, websocket_instance):
        return match_validated_event(websocket_instance, self)


class MemoryStructureControlWrapper(EventWrapper):
    model = MemoryStructureControl
    api_path: str = "structure_control_memory"
    scenario_name = "structurecontrolmemory"

    @property
    def extra_steps(self):
        wait_for_validation = WaitForMemoryStructureControlWrapper(
            data=self.instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


class WaitForFileStructureControlWrapper(WaitForEventValidation):
    model = FileStructureControl
    scenario_name = "waitforFilestructurecontrol"

    def matches(self, websocket_instance):
        return match_validated_event(websocket_instance, self)


class FileStructureControlWrapper(EventWrapper):
    model = UploadEventFile
    api_path: str = "structure_control_file"
    scenario_name = "structurecontrolfile"
    filepath: Path = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        file_controls = self.api.simulations_events_structure_control_file_list(
            simulation_pk=self.simulation.id
        )
        for file_control in file_controls.results:
            if file_control.file:
                file_control.file.filename == self.instance.filename
                bulk_control_instance = file_control
                break
        assert bulk_control_instance is not None

        wait_for_validation = WaitForFileStructureControlWrapper(
            data=bulk_control_instance.to_dict(),
            api_client=self._api_client,
            simulation=self.simulation,
        )
        return [wait_for_validation]


WRAPPERS = [
    TimedStructureControlWrapper,
    WaitForTimedStructureControlWrapper,
    TableStructureControlWrapper,
    WaitForTableStructureControlWrapper,
    MemoryStructureControlWrapper,
    WaitForMemoryStructureControlWrapper,
    FileStructureControlWrapper,
    WaitForFileStructureControlWrapper,
]



from threedi_api_client.api import ThreediApi
from threedi_cmd.schematisation_uploader.main import process
from pathlib import Path
import logging

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

# TODO: Adjust path to correct path
directory = Path("9_2D_FourCell_Rain/")

with ThreediApi("./staging.env", version='v3-beta') as client:
    schematisation, revision = process(client, directory)
    print(schematisation)
    print(revision)


from .base import LizardPostprocessingWrapper
from threedi_api_client.openapi.models import (
    BasicPostProcessing,
    DamagePostProcessing,
    ArrivalTimePostProcessing
)


class LizardBasicPostprocessingWrapper(LizardPostprocessingWrapper):
    model = BasicPostProcessing
    api_path: str = "basic"
    scenario_name = "lizardbasicpostprocessing"


class LizardDamagePostprocessingWrapper(LizardPostprocessingWrapper):
    model = DamagePostProcessing
    api_path: str = "damage"
    scenario_name = "lizarddamagepostprocessing"


class LizardArrivalPostprocessingWrapper(LizardPostprocessingWrapper):
    model = ArrivalTimePostProcessing
    api_path: str = "arrival"
    scenario_name = "lizardarrivalpostprocessing"


WRAPPERS = [
    LizardBasicPostprocessingWrapper,
    LizardDamagePostprocessingWrapper,
    LizardArrivalPostprocessingWrapper
]



from pathlib import Path
from .base import EventWrapper
from typing import Dict
from threedi_api_client.openapi.models import (
    ConstantLeakage,
    TimeseriesLeakage,
    FileTimeseriesLeakage,
    FileRasterLeakage,
    NetCDFRasterLeakage,
    NetCDFTimeseriesLeakage,
    Upload,
)

from .waitfor import WaitForProcessedFileWrapper


class ConstantLeakageWrapper(EventWrapper):
    model = ConstantLeakage
    api_path: str = "leakage_constant"
    scenario_name = "constantleakage"


class LeakageTimeseriesWrapper(EventWrapper):
    model = TimeseriesLeakage
    api_path: str = "leakage_timeseries"
    scenario_name = "timeseriesleakage"


class WaitForLeakageTimeseriesFileWrapper(WaitForProcessedFileWrapper):
    model = FileTimeseriesLeakage
    scenario_name = "waitforleakagetimeseriesfile"


class WaitForLeakageTimeseriesNetCDFWrapper(WaitForLeakageTimeseriesFileWrapper):
    model = NetCDFTimeseriesLeakage
    websocket_model_name = "NetCDFTimeseriesLeakage"
    scenario_name = "waitforleakagetimeseriesnetcdf"


class LeakageTimeseriesNetCDFWrapper(EventWrapper):
    model = Upload
    api_path: str = "leakage_timeseries_netcdf"
    scenario_name = "leakagetimeseriesnetcdf"
    filepath: Path = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        data = {
            "file": {"state": "processed", "filename": self.instance.filename},
            "timeout": 30,
        }
        wait_for_validation = WaitForLeakageTimeseriesNetCDFWrapper(
            data=data, api_client=self._api_client, simulation=self.simulation
        )
        return [wait_for_validation]


class WaitForLeakageRasterFileWrapper(WaitForProcessedFileWrapper):
    model = FileRasterLeakage
    scenario_name = "waitforleakagerasterfile"


class WaitForLeakageRasterNetCDFWrapper(WaitForLeakageRasterFileWrapper):
    model = NetCDFRasterLeakage
    websocket_model_name = "NetCDFRasterLeakage"
    scenario_name = "waitforleakagerasternetcdf"


class LeakageRasterNetCDFWrapper(EventWrapper):
    model = Upload
    api_path: str = "leakage_rasters_netcdf"
    scenario_name = "leakagerasternetcdf"
    filepath = None

    def initialize_instance(self, data: Dict):
        self.filepath = Path(data.pop("filepath"))
        super().initialize_instance(data)

    @property
    def extra_steps(self):
        data = {
            "file": {"state": "processed", "filename": self.instance.filename},
            "timeout": 30,
        }
        wait_for_validation = WaitForLeakageRasterNetCDFWrapper(
            data=data, api_client=self._api_client, simulation=self.simulation
        )
        return [wait_for_validation]


WRAPPERS = [
    ConstantLeakageWrapper,
    LeakageTimeseriesWrapper,
    LeakageTimeseriesNetCDFWrapper,
    WaitForLeakageTimeseriesNetCDFWrapper,
    LeakageRasterNetCDFWrapper,
    WaitForLeakageTimeseriesNetCDFWrapper,
]



from __future__ import annotations
from dataclasses import dataclass, field, asdict
from enum import Enum
from pathlib import Path
import sys
from typing import List
from urllib.parse import urlparse
from functools import lru_cache
import json

import click
import yaml
from rich.prompt import Prompt

from threedi_api_client.threedi_api_client import ThreediApiClient, Configuration
from threedi_api_client.openapi.exceptions import ApiException
from threedi_api_client.api import ThreediApi
from threedi_api_client.aio.threedi_api_client import (
    ThreediApiClient as AsyncThreediApiClient,
)

try:
    from threedi_api_client.auth import get_auth_token, refresh_api_key, is_token_usable
except ImportError:
    from threedi_api_client.threedi_api_client import (
        get_auth_token,
        refresh_api_key,
        is_token_usable,
    )

from threedi_cmd.console import console
from threedi_cmd.errors import ExitCodes

SCENARIO_DIR = Path(__file__).parent.parent / "scenarios"


class EndpointOption(Enum):
    localhost = "http://localhost:8000"
    staging = "https://api.staging.3di.live"
    production = "https://api.3di.live"


@lru_cache()
def get_settings(endpoint):
    return Settings(endpoint=endpoint)


@dataclass
class WebSocketSettings:
    api_base_url: str
    token: str
    host: str = field(init=False)
    proto: str = field(init=False)
    api_version: str = field(init=False)

    def __post_init__(self):
        parsed_url = urlparse(self.api_base_url)
        self.host = parsed_url.netloc
        self.proto = "wss" if parsed_url.scheme == "https" else "ws"
        self.api_version = parsed_url.path.lstrip("/")
        self.token = (
            self.token if self.token.startswith("Bearer") else f"Bearer {self.token}"
        )


@dataclass
class CachedConfig:
    username: str = ""
    access: str = ""
    refresh: str = ""
    organisation_uuid: str = ""
    result_folder: str = ""
    scenario_folder: str = ""

    @classmethod
    def load_from_file(cls, file: Path) -> CachedConfig:
        """
        Loads saved settings from 3di_config.yaml. See save method for available attributes.
        """
        try:
            with open(file, "r") as f:
                cached_data = yaml.load(f, Loader=yaml.FullLoader)
                return cls(**cached_data)
        except OSError:
            # settings file does not yet exist
            return cls()

    def save_to_file(self, file: Path) -> None:
        with open(file, "w") as f:
            yaml.dump(
                asdict(self),
                default_flow_style=False,
                stream=f,
            )


@dataclass
class ScenariosMeta:

    scenario_dir: Path = SCENARIO_DIR

    @property
    def scenarios(self) -> List:
        scenario_list = []
        scenario_files = self.scenario_dir.glob("*.yaml")
        for f in scenario_files:
            with f.open("r") as y:
                content = yaml.load_all(y, Loader=yaml.FullLoader)
                try:
                    meta = next(content)["meta"]
                    meta["file"] = f
                    scenario_list.append(meta)
                except KeyError:
                    raise AttributeError(
                        f"The scenario definition is missing the 'meta' section"
                    )
        return scenario_list


class Settings:
    """Settings that are saved (authentication) values between calls to the cli.

    The default settings are saved in the file '3di_config.yaml' which lives in
    the configuration folder of your operating system.
    """

    APP_NAME = "3di-scenario"
    CONFIG_FILE_NAME = "3di_config.yaml"

    def __init__(self, endpoint: str = "", websocket_api_version="v3-beta"):
        self._endpoint = EndpointOption[endpoint]
        self.cached_config = CachedConfig.load_from_file(self.config_file)

        if not is_token_usable(self.access):
            self.credentials_prompt(self.configuration)

        self.websocket_settings = WebSocketSettings(
            api_base_url=self.endpoint + "/" + websocket_api_version, token=self.access
        )

    @property
    def endpoint(self) -> str:
        return self._endpoint.value

    @property
    def username(self) -> str:
        # print(f"prop username {self._username}")
        return self.cached_config.username

    @username.setter
    def username(self, username: str):
        self.cached_config.username = username

    @property
    def access(self) -> str:
        return self.cached_config.access

    @access.setter
    def access(self, access: str):
        self.cached_config.access = access

    @property
    def refresh(self) -> str:
        return self.cached_config.refresh

    @refresh.setter
    def refresh(self, refresh: str):
        self.cached_config.refresh = refresh

    @property
    def organisation_uuid(self) -> str:
        return self.cached_config.organisation_uuid

    @organisation_uuid.setter
    def organisation_uuid(self, organisation_uuid: str):
        self.cached_config.organisation_uuid = organisation_uuid

    @property
    def result_folder(self) -> Path:
        return Path(self.cached_config.result_folder).resolve()

    @result_folder.setter
    def result_folder(self, result_folder: Path):
        if isinstance(result_folder, Path):
            result_folder = result_folder.resolve().as_posix()
        self.cached_config.result_folder = result_folder

    @property
    def scenario_folder(self) -> Path:
        return Path(self.cached_config.scenario_folder).resolve()

    @scenario_folder.setter
    def scenario_folder(self, scenario_folder: Path):
        if isinstance(scenario_folder, Path):
            scenario_folder = scenario_folder.resolve().as_posix()
        self.cached_config.scenario_folder = scenario_folder

    @property
    def app_dir(self) -> Path:
        _app_dir = Path(click.get_app_dir(app_name=self.APP_NAME))
        _app_dir.mkdir(parents=True, exist_ok=True)
        return _app_dir

    @property
    def config_file(self) -> Path:
        return self.app_dir / f"{self._endpoint.name}_{self.CONFIG_FILE_NAME}"

    @property
    def configuration(self) -> Configuration:
        configuration = Configuration(
            host=self.endpoint,
            username=f"{self.username}",
            api_key={
                "Authorization": f"{self.access}",
                "refresh": f"{self.refresh}",
            },
            api_key_prefix={"Authorization": "Bearer"},
        )
        configuration.refresh_api_key_hook = refresh_api_key
        if not self.username:
            self.credentials_prompt(configuration)
        return configuration

    @property
    def api_client(self) -> ThreediApiClient:
        client = ThreediApi(
            config={
                "THREEDI_API_HOST": self.endpoint,
                "THREEDI_API_USERNAME": self.username,
                "THREEDI_API_ACCESS_TOKEN": self.access,
                "THREEDI_API_REFRESH_TOKEN": self.refresh,
            },
            version="v3-beta",
        )
        return client

    @property
    def async_api_client(self) -> AsyncThreediApiClient:
        client = AsyncThreediApiClient(self.configuration)
        return client

    @property
    def scenarios(self) -> List:
        scenarios_meta = ScenariosMeta(self.scenario_folder)
        return scenarios_meta.scenarios

    def save(self) -> None:
        try:
            self.cached_config.save_to_file(self.config_file)
        except IOError as err:
            console.print(
                f"Saving cache to {self.config_file} failed with: {err}", style="error"
            )
            return

    @staticmethod
    def save_settings() -> None:
        ctx = click.get_current_context()
        ctx.obj.save()

    def credentials_prompt(self, config: Configuration) -> None:
        console.rule(
            f":key: Authentication required for {config.host}",
            characters="*",
            style="gold3",
        )
        kwargs = {}
        try:
            if self.username:
                kwargs.update({"default": self.username})
        except AttributeError:
            pass
        username = Prompt.ask("Username", **kwargs)
        password = Prompt.ask("Password", password=True)

        try:
            token = get_auth_token(username, password, config.host)
        except ValueError as err:
            console.print(f":collision: {err}", style="error")
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)
        except ApiException as err:
            body = json.loads(err.body)["detail"]
            console.print(
                f"Could not log you in. The server responded with: {body}",
                style="error",
            )
            sys.exit(ExitCodes.AUTHENTICATION_FAILED.value)

        # Update the settings
        self.username = username
        self.password = password
        self.access = token["access"]
        self.refresh = token["refresh"]
        config.api_key = {"Authorization": self.access, "refresh": self.refresh}



from threedi_api_client.versions import V3BetaApi
from threedi_api_client.openapi.exceptions import ApiException
from threedi_api_client.openapi.models import (
    Schematisation,
    SchematisationRevision,
    Upload,
)
from threedi_api_client.files import upload_file
from pathlib import Path
from hashlib import md5
from threedi_cmd.schematisation_uploader.db import ModelDB
from threedi_cmd.schematisation_uploader.yaml_converter import (
    convert_yaml,
    fill_yaml,
)
import time
import logging


logger = logging.getLogger(__name__)


def get_md5_sum(path: Path) -> str:
    with path.open("rb") as f:
        file_hash = md5()
        while chunk := f.read(8192):
            file_hash.update(chunk)
        return file_hash.hexdigest()


def get_or_create_schematisation(client: V3BetaApi, name: str) -> Schematisation:
    res = client.schematisations_list(name=name)

    if res.count == 0:
        # Create
        schematisation: Schematisation = client.schematisations_create(
            data={"name": name, "owner": "61f5a464c35044c19bc7d4b42d7f58cb"}
        )
        logger.info(f"Created schematisation {schematisation}")

    else:
        schematisation: Schematisation = res.results[0]
        logger.info(f"Retrieved schematisation {schematisation}")

    return schematisation


def get_or_create_revision(
    client: V3BetaApi,
    schematisation: Schematisation,
    directory: Path,
    valid_timeout: int = 120,
    backoff: int = 5,
) -> SchematisationRevision:
    """
    Only check if the latest revision matches

    valid_timeout => seconds
    backoff: => seconds
    """
    revision: SchematisationRevision | None
    paths = list(directory.glob("*.sqlite"))
    assert len(paths) == 1
    sqlite_path = paths[0]

    try:
        revision = client.schematisations_latest_revision(id=schematisation.id)
    except ApiException:
        revision = None

    model_db = ModelDB(sqlite_path, upgrade=True)
    raster_paths: dict[str, Path] = model_db.get_raster_filepaths(directory)

    def is_same(revision: SchematisationRevision) -> bool:
        if sorted([x.type for x in revision.rasters]) != sorted(raster_paths.keys()):
            return False

        # Check sqlite md5sum
        if revision.sqlite.file.etag != get_md5_sum(sqlite_path):
            return False

        # Check rasters md5sum
        for raster in revision.rasters:
            if raster.file.etag != get_md5_sum(raster_paths[raster.type]):
                return False

        return True

    if revision and is_same(revision):
        logger.info(f"Revision is already in sync: {revision.id}")
        return revision

    # Need to create the revision
    revision: SchematisationRevision = client.schematisations_revisions_create(
        schematisation_pk=schematisation.id, data={"empty": True}
    )
    logger.info(f"Created new revision: {revision.id}")

    # Upload sqlite
    upload: Upload = client.schematisations_revisions_sqlite_upload(
        id=revision.id,
        schematisation_pk=revision.schematisation_id,
        data={"filename": sqlite_path.name.lower()},
    )
    logger.info(f"Uploaded sqlite: {sqlite_path.name}")
    upload_file(upload.put_url, sqlite_path)
    logger.info("Sqlite uploaded")

    # Upload rasters
    for raster_type, raster_path in raster_paths.items():
        raster = client.schematisations_revisions_rasters_create(
            revision_pk=revision.id,
            schematisation_pk=revision.schematisation_id,
            data={"type": raster_type, "name": raster_path.name.lower()},
        )
        upload: Upload = client.schematisations_revisions_rasters_upload(
            id=raster.id,
            revision_pk=revision.id,
            schematisation_pk=revision.schematisation_id,
            data={"filename": raster_path.name.lower()},
        )
        logger.info(f"Uploading raster: {raster_path.name} [{raster_type}]")
        upload_file(upload.put_url, raster_path)
        logger.info(f"Uploaded raster: {raster_path.name}")

    # Upload initial-waterlevels

    # check if all files have been uploaded
    start_time = time.time()

    def all_uploaded(revision):
        if revision.sqlite.file.state != "uploaded":
            return False
        for raster in revision.rasters:
            if raster.file.state != "uploaded":
                return False
        return True

    logger.info("Waiting for all files to have status 'uploaded'")
    time.sleep(2)
    while (time.time() - start_time) < valid_timeout:
        revision = client.schematisations_revisions_read(
            id=revision.id, schematisation_pk=revision.schematisation_id
        )
        if all_uploaded(revision):
            break
        time.sleep(backoff)

    assert all_uploaded(
        revision
    ), f"Not all files have been successfully uploaded for {revision}"
    logger.info("All files have status 'uploaded'")

    logger.info("Committing...")
    # Commit revision
    client.schematisations_revisions_commit(
        id=revision.id,
        schematisation_pk=revision.schematisation_id,
        data={
            "commit_message": "autocommit new version",
        },
    )

    logger.info("Start waiting for modelchecker result...")
    # Wait until valid (within timeout)
    start_time = time.time()
    while (time.time() - start_time) < valid_timeout:
        time.sleep(backoff)
        revision = client.schematisations_revisions_read(
            id=revision.id, schematisation_pk=revision.schematisation_id
        )
        if revision.is_valid is not None:
            break

    assert revision.is_valid, f"Revision {revision} is not valid!!"
    logger.info("Modelchecker is valid, all done!")
    return revision


def process(
    client: V3BetaApi, directory: Path
) -> tuple[Schematisation, SchematisationRevision]:
    """
    directory: full path to test directory containing sqlite and rasters
        for example: Path("/repo_path_here/9_2D-FourCell_Rain/")
    """

    assert directory.is_dir(), "Given path is not a directory"

    schematisation = get_or_create_schematisation(client, directory.name)
    revision = get_or_create_revision(client, schematisation, directory)

    convert_yaml(directory / f"{directory.name}.yaml")
    fill_yaml(
        directory / f"{directory.name}_converted.yaml", schematisation.id, revision.id
    )
    return (schematisation, revision)



from uuid import uuid4
from datetime import datetime
from threedi_api_client.api import ThreediApi
from threedi_cmd.parser import ScenarioParser
from threedi_api_client.openapi.models import ThreediModel



def test_parser():
    # Note: smoke test parser (including jinja2-time template rendering)
    organisation = "61f5a464c35044c19bc7d4b42d7f58cb"
    threedimodel_id = 14
    simulation_name = "run_" + uuid4().hex

    context = {
        "threedimodel_id": threedimodel_id,
        "threedimodel": ThreediModel(
            id=1, name="test", revision="1", slug="foo", inpy_version="bar"),
        "organisation_uuid": organisation,
        "simulation_name": simulation_name,
        "datetime_now": datetime.utcnow().isoformat(),
    }

    file_path = "threedi_cmd/test_data/start_shutdown.yaml"

    config = {
        "THREEDI_API_HOST": "http://localhost:8000/",
        "THREEDI_API_USERNAME": "__key__",
        "THREEDI_API_PASSWORD": "notused",
    }

    client = ThreediApi(config=config, version="v3-beta")
    parser = ScenarioParser(file_path)
    parser.parse(client, "scenarios/", context)



import json
from asyncio import Queue
from inspect import isfunction
from typing import Dict, Optional
from threedi_api_client.openapi.api.v3_api import V3Api
from threedi_api_client.openapi.models import Schematisation, Simulation
from pathlib import Path
import logging
import requests

from rich.live import Live

from threedi_cmd.console import console
from threedi_cmd.models.errors import ApiModelError

logger = logging.getLogger(__name__)


def evaluate_lazy(value):
    if isfunction(value):
        return value()
    return value


def class_repr(data, model):
    params = ", ".join(
        [
            f"{key}={json.dumps(str(evaluate_lazy(value)))}"
            for key, value in data.items()
        ]
    )
    return f"{model.__name__}({params})"


class BaseWrapper:
    def __init__(
        self,
        data: Dict,
        api_client: V3Api,
        base_path: Path = None,
        refs: Dict = None,
    ):
        self._api_client = api_client
        self._data = data
        self._base_path = base_path
        self._refs = refs

    async def execute(self, queue: Queue):
        pass

    @property
    def extra_steps(self):
        return []


BLACKLIST = {
    "waitfor_timeout",
}


class ModelWrapper(BaseWrapper):
    api_path: str = None
    instance: object = None
    _internal_state = None
    model: object = None
    scenario_name = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initialize_instance(self._data)
        self.api: V3Api = self._api_client  # noqa

        # Add self to refs
        if self._data is not None and "_ref" in self._data:
            self._refs[self._data["_ref"]] = self.instance

    def resolve_func_name(self, suffix: str):
        func = getattr(self._api_client, f"{self.api_path}{suffix}")
        return func

    def initialize_instance(self, data: Optional[Dict] = None):
        """raises ApiModelError if initializing the instance fails"""
        new_data = {}
        if data is not None:
            new_data = {
                key: value
                for key, value in data.items()
                if not key.startswith("_") and key not in BLACKLIST
            }
        try:
            self.instance = self.model(**new_data)
        except Exception as err:
            msg = f"Initializing {self.model.__name__} failed: {err}"
            raise ApiModelError(msg)

    async def execute(self, queue):
        # By default save()
        self.save()
        for step in self.extra_steps:
            await step.execute(queue)

    def create(self):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = self.resolve_func_name("_create")
            res = func(self.instance)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def patch(self, data):
        with Live(console=console) as live:
            live.update(f"Patching {self.model.__name__}...")
            func = self.resolve_func_name("_partial_update")
            res = func(self.instance.id, data)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Patched {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def delete(self):
        with Live(console=console) as live:
            live.update(f"Deleting {self.model.__name__}...")
            func = self.resolve_func_name("_delete")
            res = func(self.instance.id)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Deleted {self.model.__name__}"
            )
        return res

    def save(self):
        res = None
        id = getattr(self.instance, "id", None)

        if id is None:
            # Create
            res = self.create()
        else:
            # Update (use patch by default)
            old_set = set(self._internal_state.to_dict().items())
            new_set = set(self.instance.to_dict().items())

            diff = new_set - old_set

            if diff != set():
                # There is something we can try to update
                res = self.patch(dict(diff))

        if res is not None:
            # Update values of self to match what has been returned
            for key, value in res.to_dict().items():
                if hasattr(self.instance, key) and value is not None:
                    setattr(self.instance, key, value)

            self._internal_state = res

    def __repr__(self):
        return class_repr(self._data, self.model)


class SimulationWrapper(ModelWrapper):
    model = Simulation
    api_path = "simulations"


class SimulationTemplateWrapper(ModelWrapper):
    model = Simulation
    api_path = "simulations_from_template"

    def initialize_instance(self, data):
        pass

    def resolve_func_name(self, suffix: str):
        func = getattr(self._api_client, f"{self.api_path}")
        return func

    def save(self):
        id = getattr(self.instance, "id", None)
        if id is not None:
            return

        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = self.resolve_func_name("")
            res = func(self._data)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(res, type(res))
        self.instance = res


class SchematisationWrapper(ModelWrapper):
    model = Schematisation
    api_path = "schematisations"


class SimulationChildWrapper(ModelWrapper):
    simulation = None
    base_path = "simulations_"

    def __init__(self, *args, **kwargs):
        self.simulation = kwargs.pop("simulation", None)
        super().__init__(*args, **kwargs)

    @property
    def instance_name(self):
        return self.scenario_name

    def resolve_func_name(self, suffix: str):
        func = getattr(self._api_client, f"{self.base_path}{self.api_path}{suffix}")
        return func

    def create(self, *args, **kwargs):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = self.resolve_func_name("_create")
            res = func(self.simulation.id, self.instance, *args, **kwargs)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def patch(self, data):
        with Live(console=console) as live:
            live.update(f"Patching {self.model}...")
            func = self.resolve_func_name("_partial_update")
            res = func(self.instance.id, self.simulation.id, data)
            live.update(f":heavy_check_mark: [bold spring_green4] Patched {self.model}")
        return res

    def delete(self):
        func = self.resolve_func_name("_delete")
        res = func(self.instance.id, self.simulation.id)
        logger.debug("Deleted %s with id: %s", self.instance_name, self.instance.id)
        return res


class SchematisationChildWrapper(ModelWrapper):
    schematisation = None
    base_path = "schematisations_"

    def __init__(self, *args, **kwargs):
        self.schematisation = kwargs.pop("schematisation", None)
        super().__init__(*args, **kwargs)

    def resolve_func_name(self, suffix: str):
        func = getattr(self._api_client, f"{self.base_path}{self.api_path}{suffix}")
        return func

    @property
    def instance_name(self):
        return self.scenario_name

    def create(self, *args, **kwargs):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = self.resolve_func_name("_create")
            res = func(self.schematisation.id, self.instance, *args, **kwargs)
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def patch(self, data):
        with Live(console=console) as live:
            live.update(f"Patching {self.model}...")
            func = self.resolve_func_name("_partial_update")
            res = func(self.instance.id, self.schematisation.id, data)
            live.update(f":heavy_check_mark: [bold spring_green4] Patched {self.model}")
        return res

    def delete(self):
        func = self.resolve_func_name("_delete")
        res = func(self.instance.id, self.schematisation.id)
        logger.debug("Deleted %s with id: %s", self.instance_name, self.instance.id)
        return res


class RevisionChildWrapper(SchematisationChildWrapper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def resolve_func_name(self, suffix: str):
        func = getattr(self._api_client, f"{self.base_path}{self.api_path}{suffix}")
        return func

    @property
    def revision(self):
        """
        Try to resolve revision based
        on _revision reference
        """
        revision_ref = self._data["_revision"]
        return self._refs[revision_ref]

    @property
    def instance_name(self):
        return self.scenario_name

    def create(self, *args, **kwargs):
        with Live(console=console) as live:
            live.update(f"Creating {self.model.__name__}...")
            func = self.resolve_func_name("_create")
            res = func(
                self.revision.id, self.schematisation.id, self.instance, *args, **kwargs
            )
            live.update(
                f":heavy_check_mark: [bold spring_green4] Created {self.model.__name__}"
            )
        console.print(self._data)
        return res

    def patch(self, data):
        with Live(console=console) as live:
            live.update(f"Patching {self.model}...")
            func = self.resolve_func_name("_partial_update")
            res = func(self.instance.id, self.schematisation.id, data)
            live.update(f":heavy_check_mark: [bold spring_green4] Patched {self.model}")
        return res

    def delete(self):
        func = self.resolve_func_name("_delete")
        res = func(self.instance.id, self.schematisation.id)
        logger.debug("Deleted %s with id: %s", self.instance_name, self.instance.id)
        return res

    def _upload_file(self, put_url=None):
        if not hasattr(self, "filepath") or not self.filepath:
            return
        if put_url is None:
            put_url = self.instance.put_url

        filepath = Path(self.filepath)
        if self._base_path:
            filepath = self._base_path / filepath
        fp = (filepath).resolve()
        # Upload file
        with fp.open("rb") as f:
            res = requests.put(put_url, data=f)
            assert res.status_code == 200

    def save(self, upload_file=True):
        super().save()

        if upload_file:
            self._upload_file()


class EventWrapper(SimulationChildWrapper):
    base_path = "simulations_events_"

    def save(self):
        super().save()

        if not hasattr(self, "filepath") or not self.filepath:
            return
        filepath = Path(self.filepath)
        if self._base_path:
            filepath = self._base_path / filepath
        fp = (filepath).resolve()

        # Upload file
        with fp.open("rb") as f:
            res = requests.put(self.instance.put_url, data=f)
            assert res.status_code == 200


class SettingsWrapper(SimulationChildWrapper):
    base_path = "simulations_settings_"


class InitialWrapper(SimulationChildWrapper):
    base_path = "simulations_initial_"


class LizardPostprocessingWrapper(SimulationChildWrapper):
    base_path = "simulations_results_post_processing_lizard_"



import yaml
import threedi_api_client.openapi.models as models

from pathlib import Path
from typing import Dict, List, Union


class Organisation:
    fields = "{{ organisation_uuid }}"


class ThreediModel:
    fields = {
        "schematisation_id": "{{ schematisation_id }}",
        "revision_id": "{{ revision_id }}",
        "auto_update": True,
    }


class Simulation:
    fields = {
        "threedimodel": ThreediModel,
        "organisation": Organisation,
        "name": "{{ simulation_name }}",
        "start_datetime": "{{ datetime_now }}",
        "duration": "{{ duration }}",
    }


scenario_mapping: Dict[str, Union[Organisation, ThreediModel, Simulation]] = {
    "organisation": Organisation,
    "simulation": Simulation,
    "threedimodel": ThreediModel,
}

event_mapping = {
    "breach": models.Breach,
    "constantlateral": models.ConstantLateral,
    "constantrain": models.ConstantRain,
    "filelateral": models.FileLateral,
    "leakagerasterlocal": models.FileRasterLeakage,
    "localrainconstant": models.ConstantLocalRain,
    "localraintimeseries": models.TimeseriesLocalRain,
    "rainrasterlizard": models.LizardRasterRain,
    "raintimeserieslizard": models.LizardTimeseriesRain,
    # rasteredit
    "timeseriesboundary": models.BoundaryCondition,
    # "timeseriesinflow": models.TimeseriesInflow,
    "timeserieslateral": models.TimeseriesLateral,
    "timeseriesleakage": models.TimeseriesLeakage,
    "timeseriesrain": models.TimeseriesRain,
    "timeseriessourcessinks": models.TimeseriesSourcesSinks,
    # saved_state
    # initial saved state
    "timeserieswind": models.TimeseriesWind,
    # "winddragcoefficients": models.WindDragCoefficients,
}


def convert_yaml(yaml_file: Path) -> None:
    with open(yaml_file, "r") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)

    meta = {}
    if "meta" in data:
        meta = data["meta"]

    scenario = {}
    scenario["steps"] = _get_events(data["scenario"]["steps"])
    del data["scenario"]["steps"]

    for key, values in data["scenario"].items():
        if key in scenario_mapping:
            model = scenario_mapping[key]
            params = _extract_scenario(model, values)
            scenario[key] = params

    test_name = yaml_file.stem
    with open(yaml_file.parent / f"{test_name}_converted.yaml", "w") as f:
        yaml.dump({"meta": meta, "scenario": scenario}, f)


def fill_yaml(yaml_file: Path, schematisation_id=None, revision_id=None) -> None:
    with open(yaml_file, "r") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)

    data["scenario"]["simulation"]["threedimodel"]["schematisation_id"] = schematisation_id
    data["scenario"]["simulation"]["threedimodel"]["revision_id"] = revision_id

    with open(yaml_file, "w") as f:
        yaml.dump(data, f)


def _openapi_to_dict(attribute_map: Dict, required_values: List, values: Dict) -> Dict:
    # Create placeholders for all required values and fill values if they are present
    # in the attribute_map
    params = {} 
    for key in required_values:
        if key in values:
            params[key] = values[key]
        else:
            params[key] = "{{ " + key + " }}"

    for key, value in values.items():
        if key in attribute_map:
            params[key] = value

    return params


def _extract_scenario(model: Union[Organisation, ThreediModel, Simulation], values: Dict):
    # Recursivlely extract fields van scenario models
    if isinstance(model.fields, str):
        return model.fields

    params = {}
    for key, v in model.fields.items():
        if key in scenario_mapping:
            params[key] = _extract_scenario(scenario_mapping[key], {})
        elif key in values:
            params[key] = values[key]
        else:
            params[key] = v

    return params

def _get_events(steps: List[Dict]) -> List[Dict]:
    events = []
    for step in steps:
        key = list(step.keys())[0]
        if key in event_mapping:
            model = event_mapping[key]
            params = _openapi_to_dict(model.attribute_map, model.required_fields, step[key])
            events += [{key: params}]

    events += [{"action": {"name": "start", "waitfor_timeout": 1800}}]
    events += [{"waitforstatus": {"name": "finished", "timeout": 1800}}]
    return events



import typer

from threedi_cmd.commands.active_simulations import active_sims_app
from threedi_cmd.commands.scenarios import scenario_app
from threedi_cmd.commands.api import api_app
from threedi_cmd.commands.suite import suite_app
from threedi_cmd.plugins.models import AppRegistry
from threedi_cmd.plugins.models import AppMeta


core = typer.Typer()

# main app definition
core = AppMeta(app=core, name="core", help="Interact with various parts of 3Di")

live = AppMeta(
    app=active_sims_app,
    name="live",
    help="Get real time updates of running simulations",
    add_to="core",
)

scenarios = AppMeta(
    app=scenario_app,
    name="scenarios",
    help="Manage your local scenarios",
    add_to="core",
)

api = AppMeta(
    app=api_app, name="api", help="Interact with with the 3Di API", add_to="core"
)

suite = AppMeta(app=suite_app, name="suite", help="Run a scenario suite", add_to="core")

registry = AppRegistry(apps={inst.name: inst for inst in [core, live, scenarios, api]})
